"""
zerokspot.recipe.git is a small recipe that allows you to use git
repositories in a similar way like iw.recipe.subversion does for
subversion repositories::

    [myapp]
    recipe = zerokspot.recipe.git
    repository = <url-of-repository>
    paths = <relative-paths-to-packages-inside-repository>
    branch = <name-of-branch> # default: "master"
    rev = <name-of-revision> # default: None
    newest = [true|false] # default: false, stay up to date even when
                          # when updating unless rev is set
    as_egg = [true|false] # default: false, install the fetched repo as
                          # egg
    cache-name = <name-in-download-cache> # default: None

This would store the cloned repository in ${buildout:directory}/parts/myapp.
"""

import subprocess
import os.path
import zc.buildout
import shutil


def git(operation, args, message, ignore_errnos=None, verbose=False):
    """
    Execute a git operation with the given arguments. If it fails, raise an
    exception with the given message. If ignore_errnos is a list of status
    codes, they will be not handled as errors if returned by git.
    """
    if verbose:
        real_args = list(args)
    else:
        real_args = ['-q'] + list(args)
    
    command = r'git %s ' + ' '.join(('"%s"', ) * len(real_args))
    command = command % ((operation, ) + tuple(real_args))
    status = subprocess.call(command, shell=True)
    if ignore_errnos is None:
        ignore_errnos = []
    if status != 0 and status not in ignore_errnos:
        raise zc.buildout.UserError(message)


def get_reponame(url, branch = None, rev = None):
    """
    Given the URL of a repository, this function returns the name of it after
    a clone process.
    """
    base = filter(lambda x: len(x), url.split('/'))[-1]
    if base.endswith('.git'):
        base = base[:-4]

    if rev != None or branch != None:
        base = base + '@' + (rev or branch)
    return base


class Recipe(object):
    """
    This recipe supports following options:

    repository
        Path to the repository that should be cloned

    branch
        Which branch should be cloned. If none is given, "master" is used by
        default.

    rev
        Revision that should be used. This is useful if you want to freeze
        the source at a given revision. If this is used, an update won't do
        all that much when executed.
        
    paths
        List of relative paths to packages to develop. Must be used together
        with as_egg=true.
        
    as_egg
        Set to True if you want the checkout to be registered as a
        development egg in your buildout.
        
    """

    def __init__(self, buildout, name, options):
        self.buildout, self.name, self.options = buildout, name, options
        self.repository = options['repository']
        self.branch = options.get('branch', 'master')
        self.rev = options.get('rev', None)
        self.newest = options.get('newest',
                buildout['buildout'].get('newest', "false")).lower() == 'true'
        self.offline = options.get('offline', 'false').lower() == 'true'
        self.cache_install = self.offline or options.get('install-from-cache',
                buildout['buildout'].get('install-from-cache', 'false')) \
                        .lower() == 'true'
        self.cache_name = options.get('cache-name',
                get_reponame(self.repository))
        self.download_cache = self.buildout['buildout'] \
                .get('download-cache', None)
        if self.download_cache:
            self.cache_path = os.path.join(
                    buildout['buildout']['download-cache'],
                    self.cache_name)
        else:
            self.cache_path = None
        options['location'] = os.path.join(
                buildout['buildout']['parts-directory'], name)
        self.as_egg = options.get('as_egg', 'false').lower() == 'true'
        self.root_dir = self.buildout['buildout']['directory']
        self.cache_created = False
        self.cache_updated = False
        self.part_updated = False
        self.cache_cloned = False
        self.installed_from_cache = False
        self.paths = options.get('paths', None)
        self.verbose = int(buildout['buildout'].get('verbosity', 0)) > 0

    def install(self):
        """
        Method called when installing a part (or when the part's config
        was changed. It clones the the given git repository and checks
        out the requested branch or commit.

        Returns the path to the part's directory.
        """
        if self.cache_install:
            if not self.download_cache:
                raise zc.buildout.UserError("Offline mode requested and no "
                                            "download-cache specified")
            if os.path.exists(self.cache_path):
                self._clone_cache()
                self.installed_from_cache = True
            else:
                raise zc.buildout.UserError("No repository in the download "
                                            "cache directory.")
        else:
            if self.download_cache:
                if not os.path.exists(self.cache_path):
                    self._clone_upstream()
                if self.newest:
                    # Update the cache first
                    self._update_cache()
                else:
                    self.installed_from_cache = True
                self._clone_cache()
            else:
                self._clone(self.repository, self.options['location'])
        if self.as_egg:
            self._install_as_egg()
        return self.options['location']

    def update(self):
        """
        Called when the buildout is called again without the local
        configuration having been altered. If no revision was
        requested and the newest-option enabled it tries to update the
        requested branch.
        """
        if self.rev is None and self.newest:
            # Do an update of the current branch
            if self.verbose:
                print "Pulling updates from origin"
            if not self.cache_install and self.download_cache:
                self._update_cache()
            self._update_part()
            os.chdir(self.options['location'])
            if self.as_egg:
                self._install_as_egg()
        else:
            # "newest" is also automatically disabled if "offline"
            # is set.
            if self.verbose:
                print "Pulling disable for this part"

    def _clone(self, from_, to):
        """
        Clone a repository located at ``from_`` to ``to``.
        """
        try:
            git('clone', (from_, to), "Couldn't clone %s into %s" % (
                    from_, to, ))
            os.chdir(to)

            if self.branch != 'master':
                if not '[branch "%s"]' % self.branch in open(os.path.join('.git', 'config')).read():
                    git('branch', ('--track', self.branch, 'origin/%s' % self.branch),
                        "Failed to set up to track remote branch", verbose=True)
                if not "ref: refs/heads/%s" % self.branch in open(os.path.join('.git', 'HEAD')).read():
                    git('checkout', (self.branch,), "Failed to switch to branch '%s'" % self.branch,
                        ignore_errnos=[128])

            if self.rev is not None:
                git('checkout', (self.rev, ), "Failed to checkout revision")
        finally:
            os.chdir(self.root_dir)

    def _clone_cache(self):
        """
        Clone the cache into the parts directory.
        """
        if not os.path.exists(self.cache_path):
            self._clone_upstream()
        self._clone(self.cache_path, self.options['location'])
        self.cache_cloned = True

    def _clone_upstream(self):
        """
        Clone the upstream repository into the cache
        """
        self._clone(self.repository, self.cache_path)
        self.cache_created = True

    def _update_cache(self):
        """
        Updates the cached repository.
        """
        self._update_repository(self.cache_path)
        self.cache_updated = True

    def _update_part(self):
        """
        Updates the repository in the buildout's parts directory.
        """
        self._update_repository(self.options['location'])
        self.part_updated = True

    def _update_repository(self, path):
        """
        Update the repository from the given path
        """
        try:
            os.chdir(path)
            git('pull', ('origin', self.branch, ),
                    "Failed to update repository")
        finally:
            os.chdir(self.root_dir)

    def _install_as_egg(self):
        """
        Install clone as development egg.
        """
        def _install(path, target):
            zc.buildout.easy_install.develop(path, target)
            
        target = self.buildout['buildout']['develop-eggs-directory']
        if self.paths:
            for path in self.paths.split():
                path = os.path.join(self.options['location'], path.strip())
                _install(path, target)
        else:
            _install(self.options['location'], target)



from __future__ import with_statement
import unittest
import sys
import os
import tempfile
from zc.buildout import testing
import zc.buildout

FILE = os.path.abspath(__file__)

def do_buildout(buildout_file, options=None):
    if options is None:
        options = []
    build = zc.buildout.buildout.Buildout(buildout_file, options)
    build.init(None)
    build.install(None)
    return build

class UtilsTests(unittest.TestCase):
    """
    Test cases for the utility functions.
    """

    def test_repo_url(self):
        """
        Tests for get_reponame.
        """

        from zerokspot.recipe.git import get_reponame
        tests = (
                'http://domain.com/test/', 
                'http://domain.com/test.git', 
                'http://domain.com/test.git/',
                'http://domain.com/test',
                '/Users/test/repos/test'
                )
        for t in tests:
            self.assertEqual('test', get_reponame(t))

    def testBranch(self):
        """
        Test reponame with branch
        """
        from zerokspot.recipe.git import get_reponame
        self.assertEqual(
            'test@cool-feature', get_reponame('http://domain.com/test.git', 'cool-feature'))

    def testRev(self):
        """
        Test reponame with revision
        """
        from zerokspot.recipe.git import get_reponame
        self.assertEqual(
            'test@1234', get_reponame('http://domain.com/test.git', rev = '1234'))

    def testBranchAndRev(self):
        """
        Test reponame with branch and revision
        """
        from zerokspot.recipe.git import get_reponame
        self.assertEqual(
            'test@1234', get_reponame('http://domain.com/test.git', 'cool-feature', '1234'))


class RecipeTests(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        os.mkdir(os.path.join(self.tempdir, 'downloads'))

        self.tempcache = tempfile.mkdtemp()
        self.temprepos = tempfile.mkdtemp()
        self.repo_name = 'testrepo'
        self.temprepo = os.path.join(self.temprepos, self.repo_name)

        testing.mkdir(self.temprepo)

        os.chdir(self.tempdir)

        testing.system('cd %s && git init' % self.temprepo)
        testing.write(self.temprepo, 'test.txt', 'TEST')
        testing.system('cd %s && git add test.txt && git commit -m "Init"' % self.temprepo)
        testing.write(self.temprepo, 'test2.txt', 'TEST')
        testing.system('cd %s && git checkout -b test && git add test2.txt && git commit -m "Test-branch" && git checkout master' % self.temprepo)

    def tearDown(self):
        testing.rmdir(self.tempdir)
        testing.rmdir(self.tempcache)
        testing.rmdir(self.temprepos)

    def _buildout(self, options=None):
        return do_buildout(os.path.join(self.tempdir, 'buildout.cfg'), options)

    def testFetch(self):
        """
        Tests if the basic cloning and updating of the repository works.
        """

        testing.write(self.tempdir, 'buildout.cfg', """
[buildout]
parts = gittest

[gittest]
recipe = zerokspot.recipe.git
repository = %(repo)s
        """ % {'repo' : self.temprepo})
        self._buildout()
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, 'parts', 'gittest', 'test.txt')))
        self._buildout()

    def testRaiseExceptionOnAbsentCache(self):
        """
        install-from-cache without cache should raise an exception
        """
        testing.write(self.tempdir, 'buildout.cfg', """
[buildout]
parts = gittest
download-cache = %(cache)s
install-from-cache = true

[gittest]
recipe = zerokspot.recipe.git
repository = %(repo)s
        """ % {'repo' : self.temprepo, 'cache': self.tempcache})

        build = zc.buildout.buildout.Buildout(
                    os.path.join(self.tempdir, 'buildout.cfg'), [])
        self.assertRaises(zc.buildout.UserError, build.install, None)


    def testOffline(self):
        """
        Tests if install from the dowload-cache works.
        """
        testing.write(self.tempdir, 'buildout.cfg', """
[buildout]
parts = gittest
download-cache = %(cache)s

[gittest]
recipe = zerokspot.recipe.git
repository = %(repo)s
        """ % {'repo' : self.temprepo, 'cache': self.tempcache})

        # First install as usual
        build = self._buildout()
        self.assertFalse(build['gittest'].recipe.installed_from_cache)

        # clear buildout
        os.unlink(os.path.join(build['buildout']['directory'], '.installed.cfg'))
        testing.rmdir(build['buildout']['directory'], 'parts')

        # now install from cache
        build = self._buildout([('buildout', 'install-from-cache', 'true')])
        self.assertTrue(build['gittest'].recipe.installed_from_cache)
        build = self._buildout([('buildout', 'install-from-cache', 'true')])

    def testNonstandardBranch(self):
        """
        Tests if install from the dowload-cache works with a non-standard branch.
        """
        testing.write(self.tempdir, 'buildout.cfg', """
[buildout]
parts = gittest
download-cache = %(cache)s

[gittest]
recipe = zerokspot.recipe.git
branch = test
repository = %(repo)s
        """ % {'repo' : self.temprepo, 'cache': self.tempcache})
        build = self._buildout()
        recipe = build['gittest'].recipe
        self.assertTrue(os.path.exists(os.path.join(recipe.cache_path, 'test2.txt')))
        self.assertTrue(os.path.exists(os.path.join(recipe.options['location'], 'test2.txt')))
        self._buildout()

    def testSingleEgg(self):
        repo = os.path.join(os.path.dirname(FILE), '../../../')
        testing.write(self.tempdir, 'buildout.cfg', """
[buildout]
parts = gittest

[gittest]
recipe = zerokspot.recipe.git
repository = %(repo)s
as_egg = true
        """ % {'repo': repo})
        buildout = self._buildout()
        installs = os.listdir(buildout['buildout']['develop-eggs-directory'])
        self.assertTrue('zerokspot.recipe.git.egg-link' in installs)

class MultiEggTests(unittest.TestCase):
    def setUp(self):
        self.projectdir = tempfile.mkdtemp()
        self.temprepo = tempfile.mkdtemp()
        for i in range(2):
            base_dir = os.path.join(self.temprepo, 'project%d' % (i, ))
            testing.mkdir(base_dir)
            setup_py = '''from setuptools import setup
setup(name='project%(num)d', version='1.0', py_modules=['test'])''' % {'num': i}
            with open(os.path.join(base_dir, 'setup.py'), 'w+') as fp:
                fp.write(setup_py)
            with open(os.path.join(base_dir, 'test.py'), 'w+') as fp:
                fp.write('# Dummy module')
        testing.system('cd %s && git init && git add * && git commit -m "Test"' % self.temprepo)

    def testBasic(self):
        testing.write(self.projectdir, 'buildout.cfg', """
[buildout]
parts = gittest

[gittest]
recipe = zerokspot.recipe.git
repository = %(repo)s
as_egg = true
paths = 
    project0
    project1
""" % {'repo': self.temprepo})
        buildout = do_buildout(os.path.join(self.projectdir, 'buildout.cfg'))
        installs = os.listdir(buildout['buildout']['develop-eggs-directory'])
        self.assertTrue('project0.egg-link' in installs)
        self.assertTrue('project1.egg-link' in installs)

    def tearDown(self):
        testing.rmdir(self.temprepo)
        testing.rmdir(self.projectdir)

if __name__ == '__main__':
    sys.path.insert(0,  os.path.normpath(
                            os.path.join(
                                os.path.dirname(__file__) or os.getcwd(),
                                '../../../'
                            )))
    unittest.main()



from setuptools import setup, find_packages

setup(
        name='zerokspot.recipe.git',
        author='Horst Gutmann',
        author_email='zerok@zerokspot.com',
        description='Simple zc.buildout recipe for sources in a git repository',
        long_description = open('README.rst', 'r').read(),
        version='0.5.5',
        url='http://github.com/zerok/zerokspot.gitrecipe/',
        install_requires=['setuptools', 'zc.buildout'],
        namespace_packages=['zerokspot'],
        packages=find_packages(exclude=['ez_setup']),
        entry_points={'zc.buildout': ['default = zerokspot.recipe.git:Recipe']},
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Plugins',
            'Framework :: Buildout',
            'Intended Audience :: Developers',
            'Programming Language :: Python',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: System :: Installation/Setup',
            'License :: OSI Approved :: BSD License',
            ]
        )
